import numpy as np
from math import sqrt


def generate_anchors(img_size, feature_map_size, aspect_ratio, scale, crop_out_of_range=True, keep_dim=False):
    """
    ```
    prior_boxes = od.anchor.generate_anchors(img_size=550,
                                             feature_map_size=[69, 35, 18, 9, 5],
                                             aspect_ratio=[1, 0.5, 2],
                                             scale=[24, 48, 96, 192, 384])

    print(prior_boxes.shape)
    ```
    :param img_size:
    :param feature_map_size:
    :param aspect_ratio:
    :param scale:
    :param crop_out_of_range:
    :param keep_dim:
    :return: (N, 4) => [x1, y1, x2, y2], if not keep_dim
             (N, fea_size, fea_size, ar_size, 4) => [x1, y1, x2, y2], if keep_dim
    """
    fea_sizes = feature_map_size
    aspect_ratios = aspect_ratio
    anchor_sizes = scale

    ar = np.array(aspect_ratios, dtype=np.float32)
    ar = np.sqrt(ar)

    prior_boxes = []
    for fea_size, anchor_size in zip(fea_sizes, anchor_sizes):
        x = np.arange(fea_size)
        y = np.arange(fea_size)
        gx, gy = np.meshgrid(x, y)

        s = float(img_size) / (fea_size + 1)

        gcx = (gx + 1) * s
        gcy = (gy + 1) * s
        gcx = np.reshape(gcx, [fea_size, fea_size, 1]).astype(np.float32)
        gcy = np.reshape(gcy, [fea_size, fea_size, 1]).astype(np.float32)

        ws = np.ones([fea_size, fea_size, len(aspect_ratios)], dtype=np.float32) * anchor_size
        hs = np.copy(ws)

        ar2 = np.repeat(ar, fea_size * fea_size)
        ar2 = np.reshape(ar2, [len(aspect_ratios), fea_size, fea_size])
        ar2 = np.transpose(ar2, [1, 2, 0])

        ws *= ar2
        hs /= ar2

        gx1 = gcx - ws / 2
        gy1 = gcy - hs / 2
        gx2 = gx1 + ws
        gy2 = gy1 + hs

        if crop_out_of_range:
            gx1 = np.maximum(gx1, 0)
            gy1 = np.maximum(gy1, 0)
            gx2 = np.minimum(gx2, img_size)
            gy2 = np.minimum(gy2, img_size)

        anchor = np.stack([gx1, gy1, gx2, gy2], axis=-1)
        prior_boxes.append(anchor)

    if not keep_dim:
        prior_boxes = [np.reshape(anchor, [-1, 4]) for anchor in prior_boxes]
        prior_boxes = np.concatenate(prior_boxes, axis=0)

    return prior_boxes


def calc_area(bboxes):
    """
    :param bboxes: (N, 4) => [x1, y1, x2, y2]
    :return: (N,)
    """
    assert bboxes.ndim == 2 and bboxes.shape[1] == 4

    x1, y1, x2, y2 = bboxes[:, 0], bboxes[:, 1], bboxes[:, 2], bboxes[:, 3]
    return (x2 - x1) * (y2 - y1)


def calc_iou(a, b):
    """
    :param a: (NA, 4) => [x1, y1, x2, y2]
    :param b: (NB, 4) => [x1, y1, x2, y2]
    :return: (NA,NB)
    """
    assert a.ndim == 2 and a.shape[1] == 4
    assert b.ndim == 2 and b.shape[1] == 4

    num_a, num_b = a.shape[0], b.shape[0]
    a = a.reshape([num_a, 1, 4]).repeat(num_b, axis=1)
    b = b.reshape([1, num_b, 4]).repeat(num_a, axis=0)

    a_x1, a_y1, a_x2, a_y2 = a[:, :, 0], a[:, :, 1], a[:, :, 2], a[:, :, 3]
    b_x1, b_y1, b_x2, b_y2 = b[:, :, 0], b[:, :, 1], b[:, :, 2], b[:, :, 3]
    a_area = (a_x2 - a_x1) * (a_y2 - a_y1)
    b_area = (b_x2 - b_x1) * (b_y2 - b_y1)

    max_x1 = np.maximum(a_x1, b_x1)
    min_x2 = np.minimum(a_x2, b_x2)
    max_y1 = np.maximum(a_y1, b_y1)
    min_y2 = np.minimum(a_y2, b_y2)

    inter_w = np.maximum(0, min_x2 - max_x1)
    inter_h = np.maximum(0, min_y2 - max_y1)
    inter = inter_w * inter_h

    return inter / (a_area + b_area - inter)


def matching(anchors, gt_bboxes, gt_labels, threshold_pos=0.5, threshold_neg=0.4):
    """
    :param anchors (num_anchors, 4) => [x1, y1, x2, y2]
    :param gt_bboxes: (num_bboxes, 4) => [x1, y1, x2, y2]
    :param gt_labels: (num_bboxes,)
    :param threshold_pos: default value is 0.5
    :param threshold_neg: default value is 0.4
    :return: matched: (num_anchors,) in [-1, 0, 1], bboxes: (num_anchors, 4) => [x1, y1, x2, y2], labels: (num_anchors,)
    """
    # calculate iou => [num_anchors, num_bboxes]
    iou = calc_iou(anchors, gt_bboxes)

    # assign the max overlap gt index for each anchor
    max_iou_for_anchors = np.max(iou, axis=-1)
    max_id_for_anchors = np.argmax(iou, axis=-1)

    # pos=1, neg=-1, ignore=0
    matched = np.zeros_like(max_iou_for_anchors, dtype=int)
    matched[max_iou_for_anchors >= threshold_pos] = 1
    matched[max_iou_for_anchors <= threshold_neg] = -1

    # gather
    labels = gt_labels[max_id_for_anchors]
    bboxes = gt_bboxes[max_id_for_anchors]
    return matched, bboxes, labels, max_id_for_anchors


def bbox2offset(anchors, bboxes):
    """
    :param anchors: (N, 4) => [x1, y1, x2, y2]
    :param bboxes: (N, 4) => [x1, y1, x2, y2]
    :return: (N, 4) => r[cx, cy, w, h]
    """
    assert anchors.ndim == 2 and anchors.shape[1] == 4
    assert bboxes.ndim == 2 and bboxes.shape[1] == 4
    assert anchors.shape[0] == bboxes.shape[0]

    # convert to center form
    w = anchors[:, 2] - anchors[:, 0]
    h = anchors[:, 3] - anchors[:, 1]
    center_anchors = np.stack([anchors[:, 0] + (w / 2), anchors[:, 1] + (h / 2), w, h], axis=-1)

    w = bboxes[:, 2] - bboxes[:, 0]
    h = bboxes[:, 3] - bboxes[:, 1]
    center_gt = np.stack([bboxes[:, 0] + (w / 2), bboxes[:, 1] + (h / 2), w, h], axis=-1)

    # calculate offset
    g_hat_cx = (center_gt[:, 0] - center_anchors[:, 0]) / center_anchors[:, 2]
    g_hat_cy = (center_gt[:, 1] - center_anchors[:, 1]) / center_anchors[:, 3]
    g_hat_w = np.log(center_anchors[:, 2] / center_gt[:, 2])
    g_hat_h = np.log(center_anchors[:, 3] / center_gt[:, 3])
    offset = np.stack([g_hat_cx, g_hat_cy, g_hat_w, g_hat_h], axis=-1)
    return offset


def offset2bbox(anchors, offset):
    """
    :param anchors: anchors: (N, 4) => [x1, y1, x2, y2]
    :param offset: (N, 4) => r[cx, cy, w, h]
    :return: (N, 4) => [x1, y1, x2, y2]
    """
    # convert to center form
    w = anchors[:, 2] - anchors[:, 0]
    h = anchors[:, 3] - anchors[:, 1]
    center_anchors = np.stack([anchors[:, 0] + (w / 2), anchors[:, 1] + (h / 2), w, h], axis=-1)

    # convert from center form
    g_hat_cx, g_hat_cy, g_hat_w, g_hat_h = offset[:, 0], offset[:, 1], offset[:, 2], offset[:, 3]
    bboxes_cx = g_hat_cx * center_anchors[:, 2] + center_anchors[:, 0]
    bboxes_cy = g_hat_cy * center_anchors[:, 3] + center_anchors[:, 1]
    bboxes_w = center_anchors[:, 2] / np.exp(g_hat_w)
    bboxes_h = center_anchors[:, 3] / np.exp(g_hat_h)

    bboxes_x = bboxes_cx - bboxes_w / 2
    bboxes_y = bboxes_cy - bboxes_h / 2
    bboxes = np.stack([bboxes_x, bboxes_y, bboxes_x + bboxes_w, bboxes_y + bboxes_h], axis=-1)
    return bboxes



import open_detection as od

# dataset = od.dataset.coco.COCODataSet("/home/killf/data/数据集/mscoco")
# while True:
#     for data in dataset:
#         print(data)

anchors = od.anchor.generate_anchors(img_size=550,
                                     feature_map_size=[69, 35, 18, 9, 5],
                                     aspect_ratio=[1, 0.5, 2],
                                     scale=[24, 48, 96, 192, 384])

reader = od.dataset.coco.train("/home/killf/data/数据集/mscoco",
                               anchors=anchors,
                               num_workers=2,
                               batch_size=4,
                               max_queue=8,
                               total_iter=100,
                               use_multiprocess_reader=True)
for data in reader():
    print(data.get("image").shape)



import open_detection as od
import numpy as np

anchors = od.anchor.generate_anchors(img_size=550,
                                     feature_map_size=[69, 35, 18, 9, 5],
                                     aspect_ratio=[1, 0.5, 2],
                                     scale=[24, 48, 96, 192, 384],keep_dim=True)

print(anchors.shape)

gt_bboxes = np.array([[10, 10, 100, 120],
                      [20, 30, 150, 170]], dtype=np.float32)
gt_labels = np.array([3, 7], dtype=np.int)

matched, bboxes, labels = od.anchor.matching(anchors, gt_bboxes, gt_labels)

offset = od.anchor.bbox2offset(anchors, bboxes)
bboxes2 = od.anchor.offset2bbox(anchors, offset)

print(np.abs(bboxes - bboxes2).max())



from pycocotools.coco import COCO
import numpy as np
import sys, os, cv2, time

from open_detection.dataset.parallel import ParallelGenerator
from open_detection.anchor import *


class COCODataSet:
    def __init__(self, data_root, kind="train2017", shuffle=False, shuffle_seed=None):
        self.data_root = data_root
        self.kind = kind
        self.shuffle = shuffle
        self.shuffle_seed = shuffle_seed

        self.images_dir = os.path.join(data_root, kind)
        if kind.startswith("test"):
            self.annotation_file = False
            self.image_files = [os.path.join(self.images_dir, filename) for filename in os.listdir(self.images_dir)]
        else:
            self.annotation_file = os.path.join(data_root, "annotations", f"instances_{kind}.json")
            self.annotation_coco = COCO(self.annotation_file)
            self.image_ids = list(self.annotation_coco.imgToAnns.keys())
            self.image_ids = self.image_ids[:10]

    def __len__(self):
        return len(self.image_ids) if self.annotation_file else len(self.image_files)

    def _load_data_item(self, img_id):
        img_info = self.annotation_coco.loadImgs([img_id])[0]
        img_file = os.path.join(self.images_dir, img_info.get("file_name"))
        image = cv2.imread(img_file)

        ann_ids = self.annotation_coco.getAnnIds([img_id])
        anns = self.annotation_coco.loadAnns(ann_ids)

        gt_labels, gt_bboxes, gt_masks, gt_is_crowd = [], [], [], []
        for ann in anns:
            x, y, w, h = ann.get("bbox")
            gt_labels.append(ann.get("category_id"))
            gt_bboxes.append([x, y, x + w, y + h])
            gt_masks.append(self.annotation_coco.annToMask(ann))
            gt_is_crowd.append(ann.get("iscrowd"))

        return {
            'image': image,
            'gt_labels': gt_labels,
            'gt_bboxes': gt_bboxes,
            'gt_masks': gt_masks,
            'gt_is_crowd': gt_is_crowd,
            'anns': anns
        }

    def __iter__(self):
        # 1.test
        if self.kind.startswith("test"):
            for file in self.image_files:
                yield cv2.imread(file)
            return

        # 2.shuffle
        if self.shuffle:
            if self.shuffle_seed is not None:
                np.random.seed(self.shuffle_seed)
            np.random.shuffle(self.image_ids)

        # 3.train or val
        for img_id in self.image_ids:
            yield self._load_data_item(img_id)


def enforce_size(img, bbox, masks, image_size=550, proto_output_size=138):
    """ Ensures that the image is the given size without distorting aspect ratio. """
    h, w, _ = img.shape
    new_h, new_w = image_size, image_size

    if h == new_h and w == new_w:
        return img, bbox, masks

    # Resize the image so that it fits within new_w, new_h
    w_prime = new_w
    h_prime = h * new_w / w

    if h_prime > new_h:
        w_prime *= new_h / h_prime
        h_prime = new_h

    w_prime = int(w_prime)
    h_prime = int(h_prime)

    # Do all the resizing
    img = cv2.resize(img, (w_prime, h_prime))

    # Act like each object is a color channel
    masks = [cv2.resize(mask, (w_prime, h_prime)) for mask in masks]

    # Scale bounding boxes (this will put them in the top left corner in the case of padding)
    dx, dy = w_prime / w, h_prime / h
    bbox = [(x * dx, y * dy, w * dx, h * dy) for (x, y, w, h) in bbox]

    # Finally, pad everything to be the new_w, new_h
    pad_dims = [(0, new_h - h_prime), (0, new_w - w_prime), (0, 0)]
    img = np.pad(img, pad_dims, 'constant', constant_values=0)

    pad_dims = [(0, new_h - h_prime), (0, new_w - w_prime)]
    masks = [np.pad(mask, pad_dims, 'constant', constant_values=0) for mask in masks]
    masks = [cv2.resize(mask, (proto_output_size, proto_output_size)) for mask in masks]

    return img, bbox, masks


def train(data_root,
          anchors,
          kind="2017",
          image_size=550,
          proto_output_size=138,
          threshold_pos=0.5,
          threshold_neg=0.4,
          batch_size=32,
          shuffle=True,
          shuffle_seed=None,
          total_iter=0,
          num_workers=7,
          max_queue=32,
          use_multiprocess_reader=True):
    def transform(data):
        image = data["image"]
        gt_labels_origin = data["gt_labels"]
        gt_bboxes_origin = data["gt_bboxes"]
        gt_masks_origin = data["gt_masks"]

        image, gt_bboxes, gt_masks = enforce_size(image, gt_bboxes_origin, gt_masks_origin, image_size=image_size,
                                                  proto_output_size=proto_output_size)

        gt_labels = np.array(gt_labels_origin, dtype=np.int)
        gt_bboxes = np.array(gt_bboxes, dtype=np.float32)

        matched, gt_bboxes, gt_labels, id_for_anchors = matching(anchors, gt_bboxes, gt_labels,
                                                                 threshold_pos=threshold_pos,
                                                                 threshold_neg=threshold_neg)
        gt_bboxes = bbox2offset(anchors, gt_bboxes)

        return {
            "image": image,
            "gt_labels_origin": gt_labels_origin,
            "gt_bboxes_origin": gt_bboxes_origin,
            "gt_masks_origin": gt_masks_origin,
            "gt_labels": gt_labels,
            "gt_bboxes": gt_bboxes,
            "gt_masks": gt_masks,
            "matched": matched,
            "id_for_anchors": id_for_anchors
        }

    def batch_reader():
        generator = COCODataSet(data_root, kind="train" + kind, shuffle=shuffle, shuffle_seed=shuffle_seed)

        batch = []
        while True:
            for data in generator:
                data = transform(data)
                batch.append(data)
                if len(batch) == batch_size:
                    yield {
                        "image": np.stack([item.get("image") for item in batch]),
                        "gt_labels_origin": [item.get("gt_labels_origin") for item in batch],
                        "gt_bboxes_origin": [item.get("gt_bboxes_origin") for item in batch],
                        "gt_masks_origin": [item.get("gt_masks_origin") for item in batch],
                        "gt_labels": np.stack([item.get("gt_labels") for item in batch]),
                        "gt_bboxes": np.stack([item.get("gt_bboxes") for item in batch]),
                        "gt_masks": [item.get("gt_masks") for item in batch],
                        "matched": np.stack([item.get("matched") for item in batch]),
                        "id_for_anchors": [item.get("id_for_anchors") for item in batch]
                    }
                    batch = []

    if not use_multiprocess_reader:
        def _reader():
            cnt = 0
            for data in batch_reader():
                cnt += 1
                yield data
                if cnt >= total_iter:
                    break

        return _reader
    else:
        if sys.platform == "win32":
            print("multiprocess is not fully compatible with Windows, "
                  "you can set --use_multiprocess_reader=False if you "
                  "are training on Windows and there are errors incured "
                  "by multiprocess.")
        print("multiprocess reader starting up, it takes a while...")

    def parallel_batch_reader():
        generator = ParallelGenerator(batch_reader(), use_multiprocessing=use_multiprocess_reader)
        cnt = 0
        try:
            generator.start(max_queue_size=max_queue, workers=num_workers)
            generator_out = None
            while True:
                while generator.is_running():
                    if not generator.queue.empty():
                        generator_out = generator.queue.get()
                        break
                    else:
                        time.sleep(0.02)
                yield generator_out
                cnt += 1
                if cnt >= total_iter:
                    generator.stop()
                    return
                generator_out = None
        except Exception as e:
            print("Exception occured in reader: {}".format(str(e)))
        finally:
            generator.stop()

    return parallel_batch_reader



import os
import sys
import signal
import time
import numpy as np
import threading
import multiprocessing

try:
    import queue
except ImportError:
    import Queue as queue


class ParallelGenerator:
    """
    Builds a queue out of a data generator.

    Args:
        generator: a generator function which endlessly yields data
        use_multiprocessing (bool): use multiprocessing if True,
            otherwise use threading.
        wait_time (float): time to sleep in-between calls to `put()`.
        random_seed (int): Initial seed for workers,
            will be incremented by one for each workers.
    """

    def __init__(self,
                 generator,
                 use_multiprocessing=False,
                 wait_time=0.05,
                 random_seed=None):
        self.wait_time = wait_time
        self._generator = generator
        self._use_multiprocessing = use_multiprocessing
        self._threads = []
        self._stop_event = None
        self.queue = None
        self._manager = None
        self.seed = random_seed

    def start(self, workers=1, max_queue_size=10):
        """
        Start worker threads which add data from the generator into the queue.

        Args:
            workers (int): number of worker threads
            max_queue_size (int): queue size
                (when full, threads could block on `put()`)
        """

        def data_generator_task():
            """
            Data generator task.
            """

            def task():
                if (self.queue is not None and
                        self.queue.qsize() < max_queue_size):
                    generator_output = next(self._generator)
                    self.queue.put((generator_output))
                else:
                    time.sleep(self.wait_time)

            if not self._use_multiprocessing:
                while not self._stop_event.is_set():
                    with self.genlock:
                        try:
                            task()
                        except Exception:
                            self._stop_event.set()
                            break
            else:
                while not self._stop_event.is_set():
                    try:
                        task()
                    except Exception:
                        self._stop_event.set()
                        break

        try:
            if self._use_multiprocessing:
                self._manager = multiprocessing.Manager()
                self.queue = self._manager.Queue(maxsize=max_queue_size)
                self._stop_event = multiprocessing.Event()
            else:
                self.genlock = threading.Lock()
                self.queue = queue.Queue()
                self._stop_event = threading.Event()
            for _ in range(workers):
                if self._use_multiprocessing:
                    # Reset random seed else all children processes
                    # share the same seed
                    np.random.seed(self.seed)
                    thread = multiprocessing.Process(target=data_generator_task)
                    thread.daemon = True
                    if self.seed is not None:
                        self.seed += 1
                else:
                    thread = threading.Thread(target=data_generator_task)
                self._threads.append(thread)
                thread.start()
        except:
            self.stop()
            raise

    def is_running(self):
        """
        Returns:
            bool: Whether the worker theads are running.
        """
        return self._stop_event is not None and not self._stop_event.is_set()

    def stop(self, timeout=None):
        """
        Stops running threads and wait for them to exit, if necessary.
        Should be called by the same thread which called `start()`.

        Args:
            timeout(int|None): maximum time to wait on `thread.join()`.
        """
        if self.is_running():
            self._stop_event.set()
        for thread in self._threads:
            if self._use_multiprocessing:
                if thread.is_alive():
                    thread.join(timeout)
            else:
                thread.join(timeout)
        if self._manager:
            self._manager.shutdown()

        self._threads = []
        self._stop_event = None
        self.queue = None

    def get(self):
        """
        Creates a generator to extract data from the queue.
        Skip the data if it is `None`.

        # Yields
            tuple of data in the queue.
        """
        while self.is_running():
            if not self.queue.empty():
                inputs = self.queue.get()
                if inputs is not None:
                    yield inputs
            else:
                time.sleep(self.wait_time)



from setuptools import setup, find_packages

setup(
    name="open_detection",
    version="0.2.1",
    author="killf",
    author_email="killf@foxmail.com",
    maintainer="killf",
    maintainer_email="killf@foxmail.com",
    contact="killf",
    contact_email="killf@foxmail.com",
    url="https://github.com/killf/open_detection",
    license="Apache2.0",
    description="A simple and powerful toolkit for object detection and instance segmentation.",
    long_description="A simple and powerful toolkit for object detection and instance segmentation.",
    packages=find_packages(),
    install_requires=["numpy"],
    python_requires='>=3.7',
    classifiers=[],
    keywords=[]
)
